/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 11/06/22.
//

#ifndef FGGL_DEBUG_IMPL_LOGGING_FMT_HPP
#define FGGL_DEBUG_IMPL_LOGGING_FMT_HPP

#include <fmt/format.h>
#include <iostream>
#include <string>
#include <string_view>
#include <cstdio>

constexpr std::string_view CERR_FMT{"[{}] {}\n"};

namespace fggl::debug {
	using FmtType = const std::string_view;

	/**
	 * Logging levels
	 */
	enum class Level {
		critical = 0,
		error = 1,
		warning = 2,
		info = 3,
		debug = 4,
		trace = 5
	};

	constexpr std::string_view level_to_string(Level level) {
		switch (level) {
		case Level::critical: return "CRITICAL";
		case Level::error: return "ERROR";
		case Level::warning: return "WARNING";
		case Level::info: return "INFO";
		case Level::debug: return "DEBUG";
		case Level::trace: return "TRACE";
		default: return "UNKNOWN";
		}
	}

	inline void vlog(const char *file, int line, fmt::string_view format, fmt::format_args args) {
		fmt::print("{}: {}", file, line);
		fmt::vprint(format, args);
	}

	template<typename S, typename... Args>
	void logf(const char *file, int line, const S &format, Args &&... args) {
		vlog(file, line, format, fmt::make_args_checked<Args...>(format, args...));
	}

	#define info_va(format, ...) \
        logf(__FILE__, __LINE__, FMT_STRING(format), __VA_ARGS__)

	template<typename ...T>
	void log(Level level, FmtType fmt, T &&...args) {
		auto fmtStr = fmt::format(fmt::runtime(fmt), args...);
		fmt::print(CERR_FMT, level_to_string(level), fmtStr);
	}

	// inlined, pre-set level versions of the log function above

	template<typename ...T>
	inline void error(FmtType fmt, T &&...args) {
		auto fmtStr = fmt::format(fmt::runtime(fmt), args...);
		fmt::print(CERR_FMT, level_to_string(Level::error), fmtStr);
	}

	template<typename ...T>
	inline void warning(FmtType fmt, T &&...args) {
		auto fmtStr = fmt::format(fmt::runtime(fmt), args...);
		fmt::print(CERR_FMT, level_to_string(Level::warning), fmtStr);
	}

	template<typename ...T>
	inline void info(FmtType fmt, T &&...args) {
		auto fmtStr = fmt::format(fmt::runtime(fmt), args...);
		fmt::print(CERR_FMT, level_to_string(Level::info), fmtStr);
	}

	template<typename ...T>
	inline void debug(FmtType fmt, T &&...args) {
		auto fmtStr = fmt::format(fmt::runtime(fmt), args...);
		fmt::print(CERR_FMT, level_to_string(Level::debug), fmtStr);
	}

	template<typename ...T>
	inline void trace(FmtType fmt, T &&...args) {
		auto fmtStr = fmt::format(fmt::runtime(fmt), args...);
		fmt::print(CERR_FMT, level_to_string(Level::trace), fmtStr);
	}

}

#endif //FGGL_DEBUG_IMPL_LOGGING_FMT_HPP
