/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_MATH_FMATH_HPP
#define FGGL_MATH_FMATH_HPP

#include <cmath>
#include "fggl/math/vector.hpp"

namespace fggl::math {

	/**
	 * A 4D floating-point vector.
	 */
	using vec4f = glm::vec4;

	/**
	 * A 3D floating-point vector.
	 */
	using vec3f = glm::vec3;

	/**
	 * A 2D floating-point vector.
	 */
	using vec2f = glm::vec2;

	constexpr static const math::vec2f VEC2_ZERO{0.0F, 0.0F};
	constexpr static const math::vec3f VEC3_ZERO{0.0F, 0.0F, 0.0F};
	constexpr static const math::vec3f VEC3_ONES{1.0F, 1.0F, 1.0F};

	/**
	 * return the remainder (modulo) of value / maximum.
	 *
	 * This will return a value in the range [0, maximum), the result will always be positive, even if passed a negative
	 * input.
	 *
	 * @param value the value to wrap.
	 * @param max the maximum value that it can take.
	 * @return the wrapped value
	 */
	inline float wrap(float value, float max) {
		return fmodf(max + fmodf(value, max), max);
	}

	/**
	 * wrap value in range [min, max)
	 *
	 * @param value the value to be tested.
	 * @param min the minimum allowable value
	 * @param max the maximum allowable value
	 */
	inline float wrap(float value, float min, float max) {
		if (min > max) {
			std::swap(min, max);
		};
		value -= min;
		float rangeSize = max - min;
		return value - (rangeSize * std::floor(value / rangeSize)) + min;
	}

	/**
	 * Ensure that value is wrapped in the range [min, max].
	 * if the value is larger than max, return max.
	 * if the value is smaller than min, return min.
	 *
	 * @param value the value to be tested.
	 * @param min the minimum allowable value
	 * @param max the maximum allowable value
	 * @return value, if it is in the range [min, max], otherwise the bound that it is outside of.
	 */
	constexpr float clamp(float value, float min, float max) {
		const float valueOrMin = value < min ? min : value;
		return valueOrMin > max ? max : valueOrMin;
	}

} // namespace fggl::math


#endif //FGGL_MATH_FMATH_HPP
