/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_MATH_TYPES_HPP
#define FGGL_MATH_TYPES_HPP

#include <tuple>
#include <iostream>

#include "fggl/math/vector.hpp"

#include <glm/ext/matrix_transform.hpp>
#include <glm/gtc/quaternion.hpp>
#include <glm/gtx/transform.hpp>
#include <glm/gtx/euler_angles.hpp>
#include <glm/gtx/quaternion.hpp>

#include "fggl/math/fmath.hpp"
#include "fggl/util/guid.hpp"

#ifndef M_PI
	#define M_PI 3.14159265358979323846
#endif

#ifndef M_PI_2
	#define M_PI_2 1.57079632679489661923
#endif

namespace fggl::math {

	constexpr float PI = M_PI;
	constexpr float HALF_PI = M_PI_2;
	constexpr float TAU = PI * 2;

	using uint8 = std::uint8_t;

	// math types (aliased for ease of use)

	/**
	 * A 4D floating-point vector.
	 */
	using vec4 = glm::vec4;


	/**
	 * A 4D signed integer vector.
	 */
	using vec4i = glm::ivec4;

	/**
	 * A 4D unsigned integer vector.
	 */
	using vec4ui = glm::ivec4;

	/**
	 * A 3D floating-point vector.
	 */
	using vec3 = glm::vec3;

	/**
	 * A 3D integer vector.
	 */
	using vec3i = glm::ivec3;

	/**
	 * A 3D unsigned integer vector.
	 */
	using vec3ui = glm::ivec3;

	/**
	 * A 2D floating-point vector
	 */
	using vec2 = glm::vec2;

	/**
	 * A 2D integer vector
	 */
	using vec2i = glm::ivec2;

	/**
	 * a 2D unsigned integer vector
	 */
	using vec2ui = glm::ivec2;

	/**
	 * A 2x2 floating-point matrix.
	 */
	using mat2 = glm::mat2;

	/**
	 * A 3x3 floating-point matrix.
	 */
	using mat3 = glm::mat3;

	/**
	 * A 4x4 floating-point matrix.
	 */
	using mat4 = glm::mat4;

	/**
	 * A quaternion.
	 */
	using quat = glm::quat;

	constexpr static const math::mat4 IDENTITY_M4{1.0F};
	constexpr static const math::quat IDENTITY_Q{1.0F, 0.0, 0.0, 0.0};

	constexpr static const math::vec3 AXIS_X{1.0F, 0.0F, 0.0F};
	constexpr static const math::vec3 AXIS_Y{0.0F, 1.0F, 0.0F};
	constexpr static const math::vec3 AXIS_Z{0.0F, 0.0F, 1.0F};

	// fastFloor from OpenSimplex2
	inline int fastFloor(double x) {
		int xi = (int) x;
		return x < xi ? xi - 1 : xi;
	}

	/**
	 * Rescale a value between [min, max] into [0, 1].
	 *
	 * @param value the value to rescale
	 * @param min the minimum value of the original range
	 * @param max the maximum value of the original range
	 * @return the rescaled value, [0, 1]
	 */
	constexpr float rescale_norm(float value, float min, float max) {
		return (value - min) / (max - min);
	}

	/**
	 * Rescale a value between [min, max] into [newMin, newMax].
	 *
	 * @param value the value to rescale
	 * @param min the minimum value of the original range
	 * @param max the maximum value of the original range
	 * @param newMin the new minimum value
	 * @param newMax the new maximum value
	 * @return the rescaled value, [newMin, newMax]
	 */
	constexpr float rescale_norm(float value, float min, float max, float newMin, float newMax) {
		return newMin + ((value - min) * (newMax - newMin)) / (max - min);
	}

	/**
	 * Rescale a normalised device-coordinate value [-1, 1] into another range.
	 *
	 * @param value the value to rescale, [-1.0, 1.0]
	 * @param newMin the new minimum value
	 * @param newMax the new maximum value
	 * @return the rescaled value, [newMin, newMax]
	 */
	constexpr float rescale_ndc(float value, float newMin, float newMax) {
		return rescale_norm(value, -1, 1, newMin, newMax);
	}

	/**
	 * Rescale a normalised value [0, 1] into another range.
	 *
	 * @param value the value to rescale
	 * @param newMin the new minimum value
	 * @param newMax the new maximum value
	 * @return the rescaled value, [newMin, newMax]
	 */
	constexpr float rescale_01(float value, float newMin, float newMax) {
		return rescale_norm(value, 0, 1, newMin, newMax);
	}

	constexpr float recale_mean(float value, float avg, float max, float min) {
		return (value - avg) / (max - min);
	}

	// reference vectors
	constexpr vec3f UP{0.0f, 1.0f, 0.0f};
	constexpr vec3f FORWARD{1.0f, 0.0f, 0.0f};
	constexpr vec3f RIGHT{0.0f, 0.0f, 1.0f};

	inline glm::mat4 modelMatrix(const vec3 offset, const quat rotation) {
		return glm::translate(glm::mat4(1.0f), offset) * glm::toMat4(rotation);
	}

	inline glm::mat4 modelMatrix(glm::vec3 offset, glm::vec3 eulerAngles) {
		return modelMatrix(offset, glm::quat(eulerAngles));
	}

	// FIXME: we have multiple definitions of rays in the codebase!
	struct Ray {
		vec3 origin;
		vec3 direction;
	};

	struct Transform {
			constexpr static const util::GUID guid = util::make_guid("Transform");

			Transform() :
				m_model(IDENTITY_M4),
				m_origin(math::VEC3_ZERO),
				m_euler(math::VEC3_ZERO),
				m_scale(math::VEC3_ONES) {
			}

			// local reference vectors
			[[nodiscard]]
			inline vec3 up() const {
				return vec4(UP, 1.0) * model();
			}

			[[nodiscard]]
			inline vec3 forward() const {
				return vec4(FORWARD, 1.0) * model();
			}

			[[nodiscard]]
			inline vec3 right() const {
				return vec4(RIGHT, 1.0) * model();
			}

			inline void translate(const vec3 change) {
				m_origin += change;
			}

			inline void origin(const vec3 pos) {
				m_origin = pos;
			}

			[[nodiscard]]
			inline vec3 origin() const {
				return m_origin;
			}

			inline void scale(const vec3 scale) {
				m_scale = scale;
			}

			[[nodiscard]]
			inline vec3 scale() const {
				return m_scale;
			}

			inline void rotate(math::vec3 axis, float radianAngle) {
				// documentation claims this is in degrees, based on experimentation this actually appears to be radians...
				auto angles = axis * radianAngle;
				m_euler += angles;
			}

			inline void rotateEuler(vec3 angles) {
				m_euler += angles;
			}

			inline void euler(vec3 angles) {
				m_euler = angles;
			}

			[[nodiscard]]
			inline glm::vec3 euler() const {
				return m_euler;
			}

			[[nodiscard]]
			inline mat4 local() const {
				const glm::mat4 transformX = glm::rotate(math::IDENTITY_M4, glm::radians(m_euler.x), AXIS_X);
				const glm::mat4 transformY = glm::rotate(math::IDENTITY_M4, glm::radians(m_euler.y), AXIS_Y);
				const glm::mat4 transformZ = glm::rotate(math::IDENTITY_M4, glm::radians(m_euler.z), AXIS_Z);

				const auto rotation = transformY * transformX * transformZ;
				return glm::translate(math::IDENTITY_M4, m_origin)
					* rotation
					* glm::scale(math::IDENTITY_M4, m_scale);
			}

			[[nodiscard]]
			inline mat4 model() const {
				return local();
			}

			inline void update(const math::mat4 &parent) {
				m_model = parent * local();
			}

			inline void lookAt(vec3 target) {
//				auto direction = m_origin - target;
				auto result = glm::lookAt(m_origin, target, math::AXIS_Y);

				math::vec3 resultAngles;
				glm::extractEulerAngleXYZ(result, resultAngles.x, resultAngles.y, resultAngles.z);
				m_euler = glm::degrees(resultAngles);
			}

		private:
			mat4 m_model; // us -> world
			vec3 m_origin;
			vec3 m_euler;
			vec3 m_scale;
	};

	inline math::mat4 calc_view_matrix(const Transform &transform) {
		return glm::lookAt(transform.origin(), transform.origin() + transform.forward(), transform.up());
	}

	inline math::mat4 calc_view_matrix(const Transform &transform, vec3 target) {
		return glm::lookAt(transform.origin(), target, transform.up());
	}

}

#endif
