/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 23/07/22.
//

#include <gtest/gtest.h>
#include "fggl/util/guid.hpp"

namespace {
	// expect blank (0-width) strings to be the offset basis
	TEST(UtilHash32, Empty) {
		auto value = fggl::util::hash_fnv1a_32("");
		EXPECT_EQ( fggl::util::FNV_OFFSET_BASIS_32, value );
	}

	TEST(UtilHash64, Empty) {
		auto value = fggl::util::hash_fnv1a_64("");
		EXPECT_EQ( fggl::util::FNV_OFFSET_BASIS_64, value );
	}

	// expect single characters to work correctly
	TEST(UtilHash32, SingleChar) {
		auto value = fggl::util::hash_fnv1a_32("a");
		EXPECT_EQ( 0xe40c292c, value );
	}

	TEST(UtilHash64, SingleChar) {
		auto value = fggl::util::hash_fnv1a_64("a");
		EXPECT_EQ( 0xaf63dc4c8601ec8c, value );
	}

	// expect hello world to known values

	TEST(UtilHash32, HelloWorld) {
		auto value = fggl::util::hash_fnv1a_32("Hello World");
		EXPECT_EQ( 0xb3902527, value );
	}

	TEST(UtilHash64, HelloWorld) {
		auto value = fggl::util::hash_fnv1a_64("Hello World");
		EXPECT_EQ( 0x3d58dee72d4e0c27, value );
	}

	// expect scoped (::) notation to work correctly

	TEST(UtilHash32, Scoped) {
		auto value = fggl::util::hash_fnv1a_32("fggl::test::scoped");
		EXPECT_EQ( 0x27fd4589, value );
	}

	TEST(UtilHash64, Scoped) {
		auto value = fggl::util::hash_fnv1a_64("fggl::test::scoped");
		EXPECT_EQ( 0xd2929655f3b0cf49, value );
	}

	// sanity checks
	TEST(UtilHash32, RepeatsAreEqual) {
		auto value = fggl::util::hash_fnv1a_32("fggl::test::scoped");
		auto value2 = fggl::util::hash_fnv1a_32("fggl::test::scoped");
		EXPECT_EQ( value, value2 );
		EXPECT_EQ( false, value < value2);
		EXPECT_EQ( false, value2 < value);
	}

	TEST(UtilHash64, RepeatsAreEqual) {
		auto value = fggl::util::hash_fnv1a_64("fggl::test::scoped");
		auto value2 = fggl::util::hash_fnv1a_64("fggl::test::scoped");
		EXPECT_EQ( value, value2 );
		EXPECT_EQ( false, value < value2);
		EXPECT_EQ( false, value2 < value);
	}

	// GUID tests
	TEST(GuidTest, GuidSuffix) {
		auto guid = fggl::util::make_guid("test");
		auto guidSuffix = "test"_fid;
		EXPECT_EQ(guid, guidSuffix);
		EXPECT_EQ( false, guid < guidSuffix);
		EXPECT_EQ( false, guidSuffix < guid);
	}

	TEST(GuidTest, OrderingLE) {
		auto guid1 = fggl::util::GUID::make(0);
		auto guid2 = fggl::util::GUID::make(1);
		EXPECT_NE(guid1, guid2);
		EXPECT_EQ( true, guid1 < guid2);
		EXPECT_EQ( false, guid2 < guid1);
	}

	TEST(GuidTest, OrderingGE) {
		auto guid1 = fggl::util::GUID::make(15);
		auto guid2 = fggl::util::GUID::make(36);
		EXPECT_NE(guid1, guid2);
		EXPECT_EQ( true, guid1 < guid2);
		EXPECT_EQ( false, guid2 < guid1);
	}

	#ifdef DEBUG
	TEST(GuidTest, GuidToString) {
		fggl::util::GUID guid = fggl::util::internString("test");
		EXPECT_EQ("test", fggl::util::guidToString(guid));
	}
	#endif

}
